/*---------------------------------------------------------------------

	Copyright ( c ) 1990 by Quantum Software Systems Limited

	$Header$

	$Source$

	$Log$
	
---------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#define CMD_OPTS	"b:f:e:d:"

typedef unsigned int  WORD;
typedef unsigned char BYTE;

#define END_CHAIN_12	0x0ff0
#define END_CHAIN_16	0xfff0

long 	fat_block;
int		fat12 = -1;
int		fat_entries;
char	*device;

static union fat_index {
		unsigned int	*w;
		unsigned char	*b;
		} fatptr;

unsigned
get_next_cluster ( clusterno, fat_offset, fatbuf )
unsigned 	clusterno;
long	 	fat_offset;
char		*fatbuf;
{
	WORD	 val;
	unsigned offset;

	if ( fat12 )
		offset = (unsigned) ( (long)( clusterno * 3L / 2L ) - fat_offset );
	else
		offset = (unsigned) ( (long)( clusterno * 2L ) - fat_offset );
	
	fatptr.b = ( BYTE * ) ( (unsigned)fatbuf + offset );
	val = ( WORD ) *( fatptr.w );

	if ( fat12 )	{
		if ( clusterno & 0x01 )
			val >>= 4;
		val &= 0x0fff;
	}

	return( val );
}
		


void
main ( argc, argv )
int		 argc;
char	*argv[];
{
    char	 *endptr;
	int		 status = 0;
	int		 opt, endcl;
	unsigned cl, clno;
	long	 fatoff;
	int		 loopctr, i, last;
	int 	 fd;
	static   char	fatbuf[1536];
	
	while( (opt=getopt( argc, argv, CMD_OPTS )) != -1 )	{
		switch( opt )	{
		
		case 'b':	fat_block = strtol(optarg, &endptr,10);
					if ( errno || (*endptr && !isspace(*endptr)) ) {
						fprintf(stderr,"-b %s : ", optarg);
						exit(1);
					}
				 	break;
		case 'f':	fat12 = ( int )strtol(optarg, &endptr,10);
					if ( errno || (*endptr && !isspace(*endptr)) ) {
						fprintf(stderr,"-f %s : ", optarg);
						exit(2);
					}
					fat12 = (fat12 == 12 ? 1 : 0) ;
					break;

		case 'e':	fat_entries = ( int )strtol(optarg, &endptr,10);
					if ( errno || (*endptr && !isspace(*endptr)) ) {
						fprintf(stderr,"-f %s : ", optarg);
						exit(2);
					}
					break;

		case 'd':	device = optarg;
					break;

		default:
					exit(3);
		}
	}
	if ( fat_block==0L || fat_entries == 0 || device == (char *)0 || fat12==-1 )	{
		fprintf( stderr, "use: dumpfat -b fat_block -e files_area_clusters -f fat12 -d device_name\n");
		fprintf( stderr, "     Get these values from the \"Dosfsys -i\" output :\n");
		fprintf( stderr, "     -b : FAT start block\n");
		fprintf( stderr, "     -e : Files area clusters\n");
		fprintf( stderr, "     -f : 12 or 16 bit fat\n");
		fprintf( stderr, "     -d : device name (ie. /dev/hd0t4)\n");
		fprintf( stderr, "     ex.: dumpfat -b 2 -e 1000 -f 16 -d /dev/hd0t4\n");
		fprintf( stderr, "          (fat starts at blk 2, dump 1000 entries, 16 bit fat, device is hd0t4)\n");
		exit(0);
	}

	printf( "fat_block=%ld\n", fat_block );
	printf( "fat_entries=%u\n", fat_entries );
	printf( "fat=%d bit\n", fat12 ? 12 : 16 );
	printf( "device=<%s>\n", device );

	fat_entries += 2;		/* for the 1st 2 reserved entries */

	fd = open( device, O_RDONLY );
	if ( fd < 0 )	{
		fprintf( stderr, "OPEN of '%s' failed :", device );
		perror("");
		exit(1);
	}
		
	loopctr = (fat12 ? 1024 : 768);
	last = 0;
	clno = 0;
	fatoff = 0;
	if ( lseek( fd, (long)((fat_block-1L) * 512L), SEEK_SET ) == -1 )
		perror("lseek(): ");

	for (; !last ; )	{
		if ( clno+loopctr >= fat_entries )	{
			last = 1;
			loopctr = fat_entries - clno;
		}
		
		if ( read( fd, fatbuf, sizeof( fatbuf ) ) != sizeof( fatbuf ) )	
			perror("read(): ");

		for ( i=0; i < loopctr; i++, clno++ )	{
			cl = get_next_cluster( clno, fatoff, fatbuf );
			if ( !(i % 5) )
				printf("\n");
			if ( fat12 )
				endcl = (cl & END_CHAIN_12) == END_CHAIN_12;
			else
				endcl = (cl & END_CHAIN_16) == END_CHAIN_16;
			printf("%04xh=<%04xh>%c ", clno, cl, endcl ? '*' : ' ' );
		}

		fatoff += sizeof( fatbuf );
	}

	printf("\n");
	close( fd );
}
		
